import unittest

import numpy as np
from geopack import geopack

from pytplot import data_exists,del_data,cdf_to_tplot,tplot,subtract,tlimit
import pyspedas
from pyspedas import time_double
from pyspedas.geopack import tt89
from pyspedas.geopack import tt96
from pyspedas.geopack import tt01
from pyspedas.geopack import tts04
from pyspedas.geopack.get_tsy_params import get_tsy_params
from pyspedas.geopack.get_w_params import get_w
from pyspedas import tinterpol, tvectot
from pytplot import join_vec, store_data, get_data, tkm2re, tplot_names

from numpy.testing import assert_array_almost_equal_nulp, assert_array_max_ulp, assert_allclose


trange = ['2015-10-16', '2015-10-17']

display=False

def get_params(model, g_variables=None):
    support_trange = [time_double(trange[0])-60*60*24, 
                      time_double(trange[1])+60*60*24]
    pyspedas.kyoto.dst(trange=support_trange)
    pyspedas.projects.omni.data(trange=trange)
    join_vec(['BX_GSE', 'BY_GSM', 'BZ_GSM'])
    if model == 't01' and g_variables is None:
        g_variables = [6.0, 10.0]
    else:
        if g_variables is not None:
            if not isinstance(g_variables, str) and not isinstance(g_variables, np.ndarray):
                g_variables = None
    return get_tsy_params('kyoto_dst',
                    'BX_GSE-BY_GSM-BZ_GSM_joined',
                    'proton_density',
                    'flow_speed',
                    model,
                    pressure_tvar='Pressure',
                    g_variables=g_variables,
                    speed=True)


class LoadGeopackIdlValidationTestCases(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """
        IDL Data has to be downloaded to perform these tests
        The SPEDAS script that creates the file: projects/themis/state/cotrans/thm_cotrans_validate.pro
        """
        from pyspedas.utilities.download import download
        from pyspedas.projects.themis.config import CONFIG

        # Testing time range
        cls.t = ['2008-03-23', '2008-03-28']

        # Testing tolerance
        cls.tol = 1e-10

        # Download tplot files
        remote_server = 'https://github.com/spedas/test_data/raw/refs/heads/main/'
        remote_name = 'analysis_tools/geopack_idl_validate.cdf'

        datafile = download(remote_file=remote_name,
                            remote_path=remote_server,
                            local_path=CONFIG['local_data_dir'],
                            no_download=False)
        if not datafile:
            # Skip tests
            raise unittest.SkipTest("Cannot download data validation file")

        del_data('*')
        filename = datafile[0]
        cdf_to_tplot(filename)
        # Input parameters
        cls.iopt=3
        cls.kp=2.0
        cls.pdyn = 2.0
        cls.dsti = -30.0
        cls.yimf = 0.0
        cls.zimf = -5.0
        cls.g1 = 6.0
        cls.g2 = 10.0
        cls.w1 = 8.0
        cls.w2 = 5.0
        cls.w3 = 9.5
        cls.w4 = 30.0
        cls.w5 = 18.5
        cls.w6 = 60.0

    def test_tilt(self):
        tilt_times,tilt_vals = get_data('bt89_tilt')
        py_tilt_vals=np.zeros(len(tilt_times))
        for idx,t in enumerate(tilt_times):
            py_tilt_vals[idx] = np.degrees(geopack.recalc(t))
        store_data('py_tilt',data={'x':tilt_times,'y':py_tilt_vals})
        idl_tilt = get_data('bt89_tilt')
        py_tilt = get_data('py_tilt')
        assert_allclose(idl_tilt.y, py_tilt.y, atol=.0003)

    def test_igrf(self):
        tt89('tha_state_pos_gsm', igrf_only=True)
        py_b = get_data('tha_state_pos_gsm_bt89')
        idl_b = get_data('bt89_igrf')
        assert_allclose(py_b.y, idl_b.y, rtol = .001, atol=1.0)

    def test_tt89(self):
        tt89('tha_state_pos_gsm')
        py_b = get_data('tha_state_pos_gsm_bt89')
        idl_b = get_data('bt89')
        assert_allclose(py_b.y, idl_b.y, rtol=.001, atol=0.5)

    def test_tt96(self):
        tv_pos=get_data('tha_state_pos_gsm')
        t1=tv_pos.times[0]
        t2=tv_pos.times[-1]
        params = np.zeros([2,10])
        params[:,0] = self.pdyn
        params[:,1] = self.dsti
        params[:,2] = self.yimf
        params[:,3] = self.zimf
        store_data('parmod',data={'x':[t1,t2],'y':params})
        tinterpol('parmod','tha_state_pos_gsm',method='nearest',newname='parmod_interp')
        tt96('tha_state_pos_gsm', parmod='parmod_interp')
        tkm2re('tha_state_pos_gsm')
        tvectot('tha_state_pos_gsm_re',join_component=True)
        py_b = get_data('tha_state_pos_gsm_bt96')
        idl_b = get_data('bt96')
        subtract('bt96','tha_state_pos_gsm_bt96','bt96_diff')
        tplot(['bt96','tha_state_pos_gsm_bt96','bt96_diff','tha_state_pos_gsm_re_tot'], display=display, save_png='t96_diffs')
        tlimit(['2007-03-23/15:00','2007-03-23/17:00'])
        tplot(['bt96','tha_state_pos_gsm_bt96','bt96_diff','tha_state_pos_gsm_re_tot'], display=display)
        tlimit(full=True)
        assert_allclose(py_b.y, idl_b.y, rtol=.001, atol=0.5)

    def test_tt96_circ(self):
        tv_pos=get_data('circle_magpoles_5re_km')
        t1=tv_pos.times[0]
        t2=tv_pos.times[-1]
        params = np.zeros([2,10])
        params[:,0] = self.pdyn
        params[:,1] = self.dsti
        params[:,2] = self.yimf
        params[:,3] = self.zimf
        store_data('parmod',data={'x':[t1,t2],'y':params})
        tinterpol('parmod','circle_magpoles_5re',method='nearest',newname='parmod_interp')
        tt96('circle_magpoles_5re_km', parmod='parmod_interp')
        #tkm2re('tha_state_pos_gsm')
        #tvectot('tha_state_pos_gsm_re',join_component=True)
        py_b = get_data('circle_magpoles_5re_km_bt96')
        idl_b = get_data('tst5re_bt96')
        subtract('tst5re_bt96','circle_magpoles_5re_km_bt96','bt96_diff')
        tplot(['tst5re_bt96','circle_magpoles_5re_km_bt96','bt96_diff'], display=display, save_png='circ_t96_diffs')
        #tlimit(['2007-03-23/15:00','2007-03-23/17:00'])
        #tplot(['bt96','tha_state_pos_gsm_bt96','bt96_diff','tha_state_pos_gsm_re_tot'], display=display)
        tlimit(full=True)
        assert_allclose(py_b.y, idl_b.y, rtol=.001, atol=0.5)


    def test_tt01(self):
        tv_pos=get_data('tha_state_pos_gsm')
        t1=tv_pos.times[0]
        t2=tv_pos.times[-1]
        params = np.zeros([2,10])
        params[:,0] = self.pdyn
        params[:,1] = self.dsti
        params[:,2] = self.yimf
        params[:,3] = self.zimf
        params[:,4] = self.g1
        params[:,5] = self.g2
        store_data('parmod',data={'x':[t1,t2],'y':params})
        tinterpol('parmod', 'tha_state_pos_gsm', method='nearest',newname='parmod_interp')
        tt01('tha_state_pos_gsm', parmod='parmod_interp')
        py_b = get_data('tha_state_pos_gsm_bt01')
        idl_b = get_data('bt01')
        subtract('bt01','tha_state_pos_gsm_bt01','bt01_diff')
        tkm2re('tha_state_pos_gsm')
        tvectot('tha_state_pos_gsm_re',join_component=True)
        tplot(['bt01','tha_state_pos_gsm_bt01','bt01_diff','tha_state_pos_gsm_re_tot'], display=display,save_png='t01_diffs')
        assert_allclose(py_b.y, idl_b.y, rtol=.001, atol=0.5)

    def test_tt01_circ(self):
        tv_pos=get_data('circle_magpoles_5re_km')
        t1=tv_pos.times[0]
        t2=tv_pos.times[-1]
        params = np.zeros([2,10])
        params[:,0] = self.pdyn
        params[:,1] = self.dsti
        params[:,2] = self.yimf
        params[:,3] = self.zimf
        params[:,4] = self.g1
        params[:,5] = self.g2
        store_data('parmod',data={'x':[t1,t2],'y':params})
        tinterpol('parmod', 'circle_magpoles_5re_km', method='nearest',newname='parmod_interp')
        tt01('circle_magpoles_5re_km', parmod='parmod_interp')
        py_b = get_data('circle_magpoles_5re_km_bt01')
        idl_b = get_data('tst5re_bt01')
        subtract('tst5re_bt01','circle_magpoles_5re_km_bt01','bt01_diff')
        #tkm2re('tha_state_pos_gsm')
        #tvectot('tha_state_pos_gsm_re',join_component=True)
        tplot(['tst5re_bt01','circle_magpoles_5re_km_bt01','bt01_diff'], display=display,save_png='circ_t01_diffs')
        assert_allclose(py_b.y, idl_b.y, rtol=.001, atol=0.5)



    def test_tts04(self):
        tv_pos=get_data('tha_state_pos_gsm')
        t1=tv_pos.times[0]
        t2=tv_pos.times[-1]
        params = np.zeros([2,10])
        params[:,0] = self.pdyn
        params[:,1] = self.dsti
        params[:,2] = self.yimf
        params[:,3] = self.zimf
        params[:,4] = self.w1
        params[:,5] = self.w2
        params[:,6] = self.w3
        params[:,7] = self.w4
        params[:,8] = self.w5
        params[:,9] = self.w6
        store_data('parmod',data={'x':[t1,t2],'y':params})
        tinterpol('parmod', 'tha_state_pos_gsm', method='nearest',newname='parmod_interp')
        tts04('tha_state_pos_gsm', parmod='parmod_interp')
        py_b = get_data('tha_state_pos_gsm_bts04')
        idl_b = get_data('bts04')
        subtract('bts04','tha_state_pos_gsm_bts04','bts04_diff')
        tplot(['bts04','tha_state_pos_gsm_bts04','bts04_diff'], display=display, save_png='ts04_diffs')
        assert_allclose(py_b.y, idl_b.y, rtol=.001, atol=0.5)

    def test_circ_tts04(self):
        tv_pos=get_data('circle_magpoles_5re_km')
        t1=tv_pos.times[0]
        t2=tv_pos.times[-1]
        params = np.zeros([2,10])
        params[:,0] = self.pdyn
        params[:,1] = self.dsti
        params[:,2] = self.yimf
        params[:,3] = self.zimf
        params[:,4] = self.w1
        params[:,5] = self.w2
        params[:,6] = self.w3
        params[:,7] = self.w4
        params[:,8] = self.w5
        params[:,9] = self.w6
        store_data('parmod',data={'x':[t1,t2],'y':params})
        tinterpol('parmod', 'circle_magpoles_5re_km', method='nearest',newname='parmod_interp')
        tts04('circle_magpoles_5re_km', parmod='parmod_interp')
        py_b = get_data('circle_magpoles_5re_km_bts04')
        idl_b = get_data('tst5re_bts04')
        subtract('tst5re_bts04','circle_magpoles_5re_km_bts04','bts04_diff')
        tplot(['tst5re_bts04','circle_magpoles_5re_km_bts04','bts04_diff'], display=display, save_png='circ_ts04_diffs')
        assert_allclose(py_b.y, idl_b.y, rtol=.001, atol=0.5)


if __name__ == '__main__':
    unittest.main()




import logging
from pytplot import store_data, options, set_units

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)


def spd_pgs_moments_tplot(moments, x=None, prefix='', suffix=''):
    """
    Creates tplot variables from moments dictionaries

    Parameters
    ----------
        moments: dict
            Dictionary containing moments values returned by moments_3d

        x: numpy.ndarray
            The x-axis (time) values

        prefix: str
            Name prefix for the output variables

        suffix: str
            Name suffix for the output variables

    Returns
    -------
    list of str
        List of tplot variables returned
    """

    if x is None:
        logging.error('Error, no x-values specified')
        return None

    if not isinstance(moments, dict):
        logging.error('Error, the "moments" variable must be a hash table containing the moments')
        return None

    units_dict = {
        'density': '1/cm^3',
        'velocity': 'km/s',
        'vthermal': 'km/s',
        'flux': '1/(cm^2-s)',
        't3': 'eV',
        'magt3': 'eV',
        'avgtemp': 'eV',
        'sc_pot': 'V',
        'eflux': 'eV/(cm^2-s)',
        'mftens': 'eV/cm^3',
        'ptens': 'eV/cm^3',
        'symm_theta': 'degrees',
        'symm_phi': 'degrees',
        'symm_ang': 'degrees',
        'magf': 'nT',
        }

    for key in moments.keys():
        store_data(prefix + key + suffix, data={'x': x, 'y': moments[key]})
        units = units_dict.get(key)
        if units is not None:
            set_units(prefix + key + suffix, units)
            options(prefix+key+suffix, 'ysubtitle', '[' + units + ']')

    options(prefix + 'velocity' + suffix, 'yrange', [-800, 800])
    # Flux is really spiky, so it's hard to set good default limits...
    options(prefix + 'flux' + suffix, 'yrange', [-1e10, 1e10])

    return [prefix + key + suffix for key in moments.keys()]



from .load import load
from pytplot import options

# This routine was originally in geotail/__init__.py, until being moved to its own file.
# Please refer to __init__.py if you need to see the revision history before it was moved.
def epic(trange=['2018-11-5', '2018-11-6'],
         datatype='k0',
         suffix='',
         prefix='',
         get_support_data=False,
         varformat=None,
         varnames=[],
         downloadonly=False,
         notplot=False,
         no_update=False,
         time_clip=False,
         force_download=False):
    """
    This function loads data from the EPIC instrument

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:

        suffix: str
            The tplot variable names will be given this suffix.  By default,
            no suffix is added.

        prefix: str
            The tplot variable names will be given this prefix.
            Default: no prefix is added

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

        force_download: bool
            Download file even if local version is more recent than server version
            Default: False

    Returns
    ----------
        List of tplot variables created.

    """

    tvars = load(instrument='epi', trange=trange, datatype=datatype, suffix=suffix, prefix=prefix, get_support_data=get_support_data,
                 varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot,
                 time_clip=time_clip, no_update=no_update, force_download=force_download)

    if tvars is None or notplot:
        return tvars

    if 'IDiffI_I' + suffix in tvars:
        options('IDiffI_I' + suffix, 'spec', True)
        options('IDiffI_I' + suffix, 'ylog', True)
        options('IDiffI_I' + suffix, 'zlog', True)

    return tvars



from .load import load


# Add the loader for MagEphem cdf files.
def magephem(trange=["2018-11-5", "2018-11-6"], probe="a", cadence="1min", coord="op77q", prefix="", suffix="", force_download=False, get_support_data=False, varformat=None, varnames=[], downloadonly=False, notplot=False, no_update=False, time_clip=False):  # 1 min or 5 min  # op77q, t89d, t89q, and ts04d
    """
    This function loads the Mag Ephemeris data for the Van Allen Probes
    The data is available from:
    https://spdf.gsfc.nasa.gov/pub/data/rbsp/rbspa/ephemeris

    Parameters
    ----------
        trange : list of str, default=['2018-11-5', '2018-11-6']
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe : str or list of str, default='a'
            Spacecraft probe name: 'a' or 'b'

        cadence : str, default='1min'
            Data cadence; options: '1min', '5min'

        coord : str, default='op77q'
            Data coordinate system. options: 'op77q', 't89d', 't89q', and 'ts04d'

        prefix : str, optional
            The tplot variable names will be given this prefix. By default, no prefix is added.

        suffix : str, optional
            Suffix for tplot variable names. By default, no suffix is added.

        force_download : bool, default=False
            Download file even if local version is more recent than server version.

        get_support_data : bool, default=False
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot. By default, only loads in data with a
            "VAR_TYPE" attribute of "data".

        varformat : str, optional
            The file variable formats to load into tplot. Wildcard character
            "*" is accepted. By default, all variables are loaded in.

        varnames: list of str, optional
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool, default=False
            Set this flag to download the CDF files, but not load them into
            tplot variables

        notplot: bool, default=False
            Return the data in hash tables instead of creating tplot variables

        no_update: bool, default=False
            If set, only load data from your local cache

        time_clip: bool, default=False
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    -------
    tvars : dict or list
        List of created tplot variables or dict of data tables if notplot is True.

    Examples
    --------
    >>> magephem_vars = pyspedas.projects.rbsp.magephem(trange=['2018-11-5/10:00', '2018-11-5/15:00'], cadence='1min', coord='op77q', time_clip=True)
    >>> tplot(['L','Lsimple','Kp','Dst'])
    """
    return load(instrument="magephem", trange=trange, probe=probe, cadence=cadence, coord=coord, prefix=prefix, suffix=suffix, force_download=force_download, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)


def magephem_ect(trange=["2018-11-5", "2018-11-6"], probe="a", cadence="1min", coord="op77q", filetype="h5", prefix="", suffix="", force_download=False, get_support_data=False, varformat=None, varnames=[], downloadonly=False, notplot=False, no_update=False, time_clip=False):  # 1 min or 5 min  # op77q, t89d, t89q, and ts04d  # h5 or txt
    """
    @Author: Xiangning Chu, 2025-04-07
    Load magnetic field line mapping ephemeris data for the Van Allen Probes from the ECT website.

    This function retrieves and processes magnetic field line mapping data from the ECT website
    (https://rbsp-ect.newmexicoconsortium.org/data_pub/), which is particularly useful for periods
    when data is not available through CDAWeb. The data includes magnetic field line mapping parameters
    such as L-shell, MLT, and magnetic coordinates in various coordinate systems.

    Parameters
    ----------
    trange : list of str, default=['2018-11-5', '2018-11-6']
        Time range of interest [starttime, endtime] with the format:
        - ['YYYY-MM-DD', 'YYYY-MM-DD'] for daily ranges
        - ['YYYY-MM-DD/hh:mm:ss', 'YYYY-MM-DD/hh:mm:ss'] for specific time ranges

    probe : str or list of str, default='a'
        Spacecraft probe identifier: 'a' for RBSP-A or 'b' for RBSP-B

    cadence : str, default='1min'
        Temporal resolution of the data:
        - '1min': 1-minute resolution
        - '5min': 5-minute resolution

    coord : str, default='op77q'
        Magnetic coordinate system for the data:
        - 'op77q': Olson-Pfitzer quiet model
        - 't89d': Tsyganenko 1989 dynamic model
        - 't89q': Tsyganenko 1989 quiet model
        - 'ts04d': Tsyganenko-Sitnov 2004 dynamic model

    filetype : str, default='h5'
        File format for data retrieval:
        - 'h5': HDF5 format
        - 'txt': ASCII text format

    prefix : str, optional
        Prefix to be added to tplot variable names

    suffix : str, optional
        Suffix to be added to tplot variable names

    force_download : bool, default=False
        If True, forces download even if local version is more recent

    get_support_data : bool, default=False
        If True, loads support data (VAR_TYPE="support_data") into tplot

    varformat : str, optional
        File variable format to load into tplot (wildcard "*" accepted)

    varnames : list of str, optional
        Specific variable names to load (if empty, loads all variables)

    downloadonly : bool, default=False
        If True, downloads files without loading into tplot variables

    notplot : bool, default=False
        If True, returns data in hash tables instead of tplot variables

    no_update : bool, default=False
        If True, only loads data from local cache

    time_clip : bool, default=False
        If True, clips variables to exact time range specified in trange

    Returns
    -------
    tvars : dict or list
        List of created tplot variables or dict of data tables if notplot is True

    Examples
    --------
    >>> # Load 1-minute resolution data in Olson-Pfitzer quiet coordinates
    >>> magephem_vars = pyspedas.projects.rbsp.magephem_ect(
    ...     trange=['2018-11-5/10:00', '2018-11-5/15:00'],
    ...     cadence='1min',
    ...     coord='op77q',
    ...     time_clip=True
    ... )
    >>> # Plot L-shell and magnetic coordinates
    >>> tplot(['L', 'Lsimple', 'Kp', 'Dst'])

    Notes
    -----
    - This function is particularly useful for periods when data is not available through CDAWeb
    - The data includes magnetic field line mapping parameters essential for radiation belt studies
    - Coordinate systems are based on different magnetic field models for various geomagnetic conditions
    """
    from pyspedas.utilities.dailynames import dailynames
    from pyspedas.utilities.download import download
    from pytplot import time_clip as tclip

    from .config import CONFIG_ECT
    from .magephem_read import magephem_read_h5, magephem_read_txt

    if not isinstance(probe, list):
        probe = [probe]

    out_files = []

    if notplot:
        tvars = {}
    else:
        tvars = []

    if prefix is None:
        prefix = ""

    if suffix is None:
        suffix = ""

    for prb in probe:

        remote_path = CONFIG_ECT["remote_data_dir"]
        local_path = CONFIG_ECT["local_data_dir"]

        pathformat = "rbsp" + prb + "/MagEphem/definitive" + "/%Y/rbsp" + prb + "_def_MagEphem_" + coord.upper() + "_%Y%m%d_v*." + filetype

        # find the full remote path names using the trange
        remote_names = dailynames(file_format=pathformat, trange=trange)
        files = download(remote_file=remote_names, remote_path=remote_path, local_path=local_path, no_download=no_update, force_download=force_download)
        if files:
            out_files.extend(files)

        if not downloadonly:
            if filetype == "h5":
                tvars_o = magephem_read_h5(sorted(files), varnames=varnames, notplot=notplot, prefix=prefix, suffix=suffix)
            elif filetype == "txt":
                tvars_o = magephem_read_txt(sorted(files), varnames=varnames, notplot=notplot, prefix=prefix, suffix=suffix)

            if notplot:
                tvars.update(tvars_o)
            else:
                tvars.extend(tvars_o)

    if downloadonly:
        return sorted(out_files)

    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix="")

    return tvars



"""
This module contains the functions for reading the magephem data from the txt and h5 files from RBSP ECT website.
https://rbsp-ect.newmexicoconsortium.org/data_pub/
@Author: Xiangning Chu, 2025-04-07
"""

import linecache

import h5py
import numpy as np
import pandas as pd
from pytplot import store_data

name_332 = [
    "Time",
    "Date",
    "DOY",
    "UTC",
    "JulianDate",
    "GpsTime",
    "DipoleTiltAngle",
    "InOut",
    "Orbnumber",
    "Xgeo",
    "Ygeo",
    "Zgeo",
    "Rgeod_Lat",
    "Rgeod_Lon",
    "Rgeod_Height",
    "Rgsm_x",
    "Rgsm_y",
    "Rgsm_z",
    "Rsm_x",
    "Rsm_y",
    "Rsm_z",
    "Rgei_x",
    "Rgei_y",
    "Rgei_z",
    "Rgse_x",
    "Rgse_y",
    "Rgse_z",
    "CDMAG_MLAT",
    "CDMAG_MLON",
    "CDMAG_MLT",
    "CDMAG_R",
    "EDMAG_MLAT",
    "EDMAG_MLON",
    "EDMAG_MLT",
    "EDMAG_R",
    "IntModel",
    "ExtModel",
    "Kp",
    "Dst",
    "Bsc_gsm_x",
    "Bsc_gsm_y",
    "Bsc_gsm_z",
    "Bsc_gsm_mag",
    "FieldLineType",
    "S_sc_to_pfn",
    "S_sc_to_pfs",
    "S_pfs_to_Bmin",
    "S_Bmin_to_sc",
    "S_total",
    "d2B_ds2",
    "Sb0",
    "RadiusOfCurv",
    "Pfn_gsm_x",
    "Pfn_gsm_y",
    "Pfn_gsm_z",
    "Pfn_geo_x",
    "Pfn_geo_y",
    "Pfn_geo_z",
    "Pfn_geod_Lat",
    "Pfn_geod_Lon",
    "Pfn_geod_Height",
    "Pfn_CD_MLAT",
    "Pfn_CD_MLON",
    "Pfn_CD_MLT",
    "Pfn_ED_MLAT",
    "Pfn_ED_MLON",
    "Pfn_ED_MLT",
    "Bfn_geo_x",
    "Bfn_geo_y",
    "Bfn_geo_z",
    "Bfn_geo_mag",
    "Bfn_gsm_x",
    "Bfn_gsm_y",
    "Bfn_gsm_z",
    "Bfn_gsm_mag",
    "Loss_Cone_Alpha_n",
    "Pfs_gsm_x",
    "Pfs_gsm_y",
    "Pfs_gsm_z",
    "Pfs_geo_x",
    "Pfs_geo_y",
    "Pfs_geo_z",
    "Pfs_geod_Lat",
    "Pfs_geod_Lon",
    "Pfs_geod_Height",
    "Pfs_CD_MLAT",
    "Pfs_CD_MLON",
    "Pfs_CD_MLT",
    "Pfs_ED_MLAT",
    "Pfs_ED_MLON",
    "Pfs_ED_MLT",
    "Bfs_geo_x",
    "Bfs_geo_y",
    "Bfs_geo_z",
    "Bfs_geo_mag",
    "Bfs_gsm_x",
    "Bfs_gsm_y",
    "Bfs_gsm_z",
    "Bfs_gsm_mag",
    "Loss_Cone_Alpha_s",
    "Pmin_gsm_x",
    "Pmin_gsm_y",
    "Pmin_gsm_z",
    "Bmin_gsm_x",
    "Bmin_gsm_y",
    "Bmin_gsm_z",
    "Bmin_gsm_mag",
    "Lsimple",
    "InvLat",
    "Lm_eq",
    "InvLat_eq",
    "BoverBeq",
    "MlatFromBoverBeq",
    "M_used",
    "M_ref",
    "M_igrf",
    "Hmin_90",
    "Hmin_85",
    "Hmin_80",
    "Hmin_75",
    "Hmin_70",
    "Hmin_65",
    "Hmin_60",
    "Hmin_55",
    "Hmin_50",
    "Hmin_45",
    "Hmin_40",
    "Hmin_35",
    "Hmin_30",
    "Hmin_25",
    "Hmin_20",
    "Hmin_15",
    "Hmin_10",
    "Hmin_5",
    "Hmin_GeodLat_90",
    "Hmin_GeodLat_85",
    "Hmin_GeodLat_80",
    "Hmin_GeodLat_75",
    "Hmin_GeodLat_70",
    "Hmin_GeodLat_65",
    "Hmin_GeodLat_60",
    "Hmin_GeodLat_55",
    "Hmin_GeodLat_50",
    "Hmin_GeodLat_45",
    "Hmin_GeodLat_40",
    "Hmin_GeodLat_35",
    "Hmin_GeodLat_30",
    "Hmin_GeodLat_25",
    "Hmin_GeodLat_20",
    "Hmin_GeodLat_15",
    "Hmin_GeodLat_10",
    "Hmin_GeodLat_5",
    "Hmin_GeodLon_90",
    "Hmin_GeodLon_85",
    "Hmin_GeodLon_80",
    "Hmin_GeodLon_75",
    "Hmin_GeodLon_70",
    "Hmin_GeodLon_65",
    "Hmin_GeodLon_60",
    "Hmin_GeodLon_55",
    "Hmin_GeodLon_50",
    "Hmin_GeodLon_45",
    "Hmin_GeodLon_40",
    "Hmin_GeodLon_35",
    "Hmin_GeodLon_30",
    "Hmin_GeodLon_25",
    "Hmin_GeodLon_20",
    "Hmin_GeodLon_15",
    "Hmin_GeodLon_10",
    "Hmin_GeodLon_5",
    "Lstar_90",
    "Lstar_85",
    "Lstar_80",
    "Lstar_75",
    "Lstar_70",
    "Lstar_65",
    "Lstar_60",
    "Lstar_55",
    "Lstar_50",
    "Lstar_45",
    "Lstar_40",
    "Lstar_35",
    "Lstar_30",
    "Lstar_25",
    "Lstar_20",
    "Lstar_15",
    "Lstar_10",
    "Lstar_5",
    "DSType_90",
    "DSType_85",
    "DSType_80",
    "DSType_75",
    "DSType_70",
    "DSType_65",
    "DSType_60",
    "DSType_55",
    "DSType_50",
    "DSType_45",
    "DSType_40",
    "DSType_35",
    "DSType_30",
    "DSType_25",
    "DSType_20",
    "DSType_15",
    "DSType_10",
    "DSType_5",
    "L_90",
    "L_85",
    "L_80",
    "L_75",
    "L_70",
    "L_65",
    "L_60",
    "L_55",
    "L_50",
    "L_45",
    "L_40",
    "L_35",
    "L_30",
    "L_25",
    "L_20",
    "L_15",
    "L_10",
    "L_5",
    "Bm_90",
    "Bm_85",
    "Bm_80",
    "Bm_75",
    "Bm_70",
    "Bm_65",
    "Bm_60",
    "Bm_55",
    "Bm_50",
    "Bm_45",
    "Bm_40",
    "Bm_35",
    "Bm_30",
    "Bm_25",
    "Bm_20",
    "Bm_15",
    "Bm_10",
    "Bm_5",
    "I_90",
    "I_85",
    "I_80",
    "I_75",
    "I_70",
    "I_65",
    "I_60",
    "I_55",
    "I_50",
    "I_45",
    "I_40",
    "I_35",
    "I_30",
    "I_25",
    "I_20",
    "I_15",
    "I_10",
    "I_5",
    "K_90",
    "K_85",
    "K_80",
    "K_75",
    "K_70",
    "K_65",
    "K_60",
    "K_55",
    "K_50",
    "K_45",
    "K_40",
    "K_35",
    "K_30",
    "K_25",
    "K_20",
    "K_15",
    "K_10",
    "K_5",
    "Sb_90",
    "Sb_85",
    "Sb_80",
    "Sb_75",
    "Sb_70",
    "Sb_65",
    "Sb_60",
    "Sb_55",
    "Sb_50",
    "Sb_45",
    "Sb_40",
    "Sb_35",
    "Sb_30",
    "Sb_25",
    "Sb_20",
    "Sb_15",
    "Sb_10",
    "Sb_5",
    "Tb_90",
    "Tb_85",
    "Tb_80",
    "Tb_75",
    "Tb_70",
    "Tb_65",
    "Tb_60",
    "Tb_55",
    "Tb_50",
    "Tb_45",
    "Tb_40",
    "Tb_35",
    "Tb_30",
    "Tb_25",
    "Tb_20",
    "Tb_15",
    "Tb_10",
    "Tb_5",
    "Kappa_90",
    "Kappa_85",
    "Kappa_80",
    "Kappa_75",
    "Kappa_70",
    "Kappa_65",
    "Kappa_60",
    "Kappa_55",
    "Kappa_50",
    "Kappa_45",
    "Kappa_40",
    "Kappa_35",
    "Kappa_30",
    "Kappa_25",
    "Kappa_20",
    "Kappa_15",
    "Kappa_10",
    "Kappa_5",
]

name_278 = [
    "Time",
    "Date",
    "DOY",
    "UTC",
    "JulianDate",
    "GpsTime",
    "DipoleTiltAngle",
    "InOut",
    "Orbnumber",
    "Xgeo",
    "Ygeo",
    "Zgeo",
    "Rgeod_Lat",
    "Rgeod_Lon",
    "Rgeod_Height",
    "Rgsm_x",
    "Rgsm_y",
    "Rgsm_z",
    "Rsm_x",
    "Rsm_y",
    "Rsm_z",
    "Rgei_x",
    "Rgei_y",
    "Rgei_z",
    "Rgse_x",
    "Rgse_y",
    "Rgse_z",
    "CDMAG_MLAT",
    "CDMAG_MLON",
    "CDMAG_MLT",
    "CDMAG_R",
    "EDMAG_MLAT",
    "EDMAG_MLON",
    "EDMAG_MLT",
    "EDMAG_R",
    "IntModel",
    "ExtModel",
    "Kp",
    "Dst",
    "Bsc_gsm_x",
    "Bsc_gsm_y",
    "Bsc_gsm_z",
    "Bsc_gsm_mag",
    "FieldLineType",
    "S_sc_to_pfn",
    "S_sc_to_pfs",
    "S_pfs_to_Bmin",
    "S_Bmin_to_sc",
    "S_total",
    "d2B_ds2",
    "Sb0",
    "RadiusOfCurv",
    "Pfn_gsm_x",
    "Pfn_gsm_y",
    "Pfn_gsm_z",
    "Pfn_geo_x",
    "Pfn_geo_y",
    "Pfn_geo_z",
    "Pfn_geod_Lat",
    "Pfn_geod_Lon",
    "Pfn_geod_Height",
    "Pfn_CD_MLAT",
    "Pfn_CD_MLON",
    "Pfn_CD_MLT",
    "Pfn_ED_MLAT",
    "Pfn_ED_MLON",
    "Pfn_ED_MLT",
    "Bfn_geo_x",
    "Bfn_geo_y",
    "Bfn_geo_z",
    "Bfn_geo_mag",
    "Bfn_gsm_x",
    "Bfn_gsm_y",
    "Bfn_gsm_z",
    "Bfn_gsm_mag",
    "Loss_Cone_Alpha_n",
    "Pfs_gsm_x",
    "Pfs_gsm_y",
    "Pfs_gsm_z",
    "Pfs_geo_x",
    "Pfs_geo_y",
    "Pfs_geo_z",
    "Pfs_geod_Lat",
    "Pfs_geod_Lon",
    "Pfs_geod_Height",
    "Pfs_CD_MLAT",
    "Pfs_CD_MLON",
    "Pfs_CD_MLT",
    "Pfs_ED_MLAT",
    "Pfs_ED_MLON",
    "Pfs_ED_MLT",
    "Bfs_geo_x",
    "Bfs_geo_y",
    "Bfs_geo_z",
    "Bfs_geo_mag",
    "Bfs_gsm_x",
    "Bfs_gsm_y",
    "Bfs_gsm_z",
    "Bfs_gsm_mag",
    "Loss_Cone_Alpha_s",
    "Pmin_gsm_x",
    "Pmin_gsm_y",
    "Pmin_gsm_z",
    "Bmin_gsm_x",
    "Bmin_gsm_y",
    "Bmin_gsm_z",
    "Bmin_gsm_mag",
    "Lsimple",
    "InvLat",
    "Lm_eq",
    "InvLat_eq",
    "BoverBeq",
    "MlatFromBoverBeq",
    "M_used",
    "M_ref",
    "M_igrf",
    "Lstar_90",
    "Lstar_85",
    "Lstar_80",
    "Lstar_75",
    "Lstar_70",
    "Lstar_65",
    "Lstar_60",
    "Lstar_55",
    "Lstar_50",
    "Lstar_45",
    "Lstar_40",
    "Lstar_35",
    "Lstar_30",
    "Lstar_25",
    "Lstar_20",
    "Lstar_15",
    "Lstar_10",
    "Lstar_5",
    "DSType_90",
    "DSType_85",
    "DSType_80",
    "DSType_75",
    "DSType_70",
    "DSType_65",
    "DSType_60",
    "DSType_55",
    "DSType_50",
    "DSType_45",
    "DSType_40",
    "DSType_35",
    "DSType_30",
    "DSType_25",
    "DSType_20",
    "DSType_15",
    "DSType_10",
    "DSType_5",
    "L_90",
    "L_85",
    "L_80",
    "L_75",
    "L_70",
    "L_65",
    "L_60",
    "L_55",
    "L_50",
    "L_45",
    "L_40",
    "L_35",
    "L_30",
    "L_25",
    "L_20",
    "L_15",
    "L_10",
    "L_5",
    "Bm_90",
    "Bm_85",
    "Bm_80",
    "Bm_75",
    "Bm_70",
    "Bm_65",
    "Bm_60",
    "Bm_55",
    "Bm_50",
    "Bm_45",
    "Bm_40",
    "Bm_35",
    "Bm_30",
    "Bm_25",
    "Bm_20",
    "Bm_15",
    "Bm_10",
    "Bm_5",
    "I_90",
    "I_85",
    "I_80",
    "I_75",
    "I_70",
    "I_65",
    "I_60",
    "I_55",
    "I_50",
    "I_45",
    "I_40",
    "I_35",
    "I_30",
    "I_25",
    "I_20",
    "I_15",
    "I_10",
    "I_5",
    "K_90",
    "K_85",
    "K_80",
    "K_75",
    "K_70",
    "K_65",
    "K_60",
    "K_55",
    "K_50",
    "K_45",
    "K_40",
    "K_35",
    "K_30",
    "K_25",
    "K_20",
    "K_15",
    "K_10",
    "K_5",
    "Sb_90",
    "Sb_85",
    "Sb_80",
    "Sb_75",
    "Sb_70",
    "Sb_65",
    "Sb_60",
    "Sb_55",
    "Sb_50",
    "Sb_45",
    "Sb_40",
    "Sb_35",
    "Sb_30",
    "Sb_25",
    "Sb_20",
    "Sb_15",
    "Sb_10",
    "Sb_5",
    "Tb_90",
    "Tb_85",
    "Tb_80",
    "Tb_75",
    "Tb_70",
    "Tb_65",
    "Tb_60",
    "Tb_55",
    "Tb_50",
    "Tb_45",
    "Tb_40",
    "Tb_35",
    "Tb_30",
    "Tb_25",
    "Tb_20",
    "Tb_15",
    "Tb_10",
    "Tb_5",
    "Kappa_90",
    "Kappa_85",
    "Kappa_80",
    "Kappa_75",
    "Kappa_70",
    "Kappa_65",
    "Kappa_60",
    "Kappa_55",
    "Kappa_50",
    "Kappa_45",
    "Kappa_40",
    "Kappa_35",
    "Kappa_30",
    "Kappa_25",
    "Kappa_20",
    "Kappa_15",
    "Kappa_10",
    "Kappa_5",
]

h5_mag_name = [
    "Alpha",
    "ApogeePosGeod",
    "ApogeeTimes",
    "AscendPosGeod",
    "AscendTimes",
    "Bfn_geo",
    "Bfn_gsm",
    "Bfs_geo",
    "Bfs_gsm",
    "Bm",
    "Bmin_gsm",
    "BoverBeq",
    "Bsc_gsm",
    "CDMAG_MLAT",
    "CDMAG_MLON",
    "CDMAG_MLT",
    "CDMAG_R",
    "Date",
    "DipoleTiltAngle",
    "Doy",
    "DriftShellType",
    "Dst",
    "EDMAG_MLAT",
    "EDMAG_MLON",
    "EDMAG_MLT",
    "EDMAG_R",
    "ExtModel",
    "FieldLineType",
    "GpsTime",
    "I",
    "InOut",
    "IntModel",
    "InvLat",
    "InvLat_eq",
    "IsoTime",
    "JulianDate",
    "K",
    "Kappa",
    "Kp",
    "L",
    "Lm_eq",
    "Loss_Cone_Alpha_n",
    "Loss_Cone_Alpha_s",
    "Lsimple",
    "Lstar",
    "M_igrf",
    "M_ref",
    "M_used",
    "MlatFromBoverBeq",
    "OrbitNumber",
    "PerigeePosGeod",
    "PerigeeTimes",
    "Pfn_CD_MLAT",
    "Pfn_CD_MLON",
    "Pfn_CD_MLT",
    "Pfn_ED_MLAT",
    "Pfn_ED_MLON",
    "Pfn_ED_MLT",
    "Pfn_geo",
    "Pfn_geod_Height",
    "Pfn_geod_LatLon",
    "Pfn_gsm",
    "Pfs_CD_MLAT",
    "Pfs_CD_MLON",
    "Pfs_CD_MLT",
    "Pfs_ED_MLAT",
    "Pfs_ED_MLON",
    "Pfs_ED_MLT",
    "Pfs_geo",
    "Pfs_geod_Height",
    "Pfs_geod_LatLon",
    "Pfs_gsm",
    "Pmin_gsm",
    "RadiusOfCurv",
    "Rgei",
    "Rgeo",
    "Rgeod_Height",
    "Rgeod_LatLon",
    "Rgse",
    "Rgsm",
    "Rsm",
    "S_Bmin_to_sc",
    "S_pfs_to_Bmin",
    "S_sc_to_pfn",
    "S_sc_to_pfs",
    "S_total",
    "Sb",
    "Sb0",
    "Tb",
    "UTC",
    "d2B_ds2",
]


def getstartline(file_name):
    """
    Find the line number of the header.
    """

    with open(file_name, "rb") as f:
        line_total_num = sum(1 for line in f)

    start_line = 0
    for i in range(920, line_total_num):
        line_data = linecache.getline(file_name, i)

        # get the first 10
        if line_data[0:10] == "#  +------":
            start_line = i
            break
    if start_line == 0:
        raise ValueError("the txt file is wrong")
    return start_line


def magephem_read_txt(out_files, varnames=None, notplot=False, prefix="", suffix=""):
    """
    Read the txt file and return the data as a dictionary or save to tplot variables.
    """

    if (varnames is None) or (len(varnames) == 0):
        varnames2load = ["Time", "Lsimple", "Lm_eq", "CDMAG_MLAT", "CDMAG_MLON", "CDMAG_MLT"]
    else:
        varnames2load = varnames

    if "Time" not in varnames2load:
        varnames2load.append("Time")

    data_all = []

    for ifile, file in enumerate(out_files):
        start_line = getstartline(file)
        # Read the data using numpy
        file_pd = pd.read_csv(file, skiprows=(start_line + 2), sep=r"\s+", header=None)

        # If the time in the last row is 00:00:00, which duplicates the start of the next day, remove the last row
        if (ifile != len(out_files) - 1) and ("T00:00:00.0000Z" in file_pd.iloc[-1, 0]):
            file_pd = file_pd.iloc[:-1, :]

        if len(file_pd.shape) == 2:
            if file_pd.shape[1] == 278:
                file_pd.columns = name_278
            elif file_pd.shape[1] == 332:
                file_pd.columns = name_332
            else:
                raise ValueError(f"This date file has some problem with the number of columns {file_pd.shape}!! \n {file}")
        else:
            raise ValueError(f"This date file has some problem with only one column!! \n {file}")

        # choose the columns that
        for name in varnames2load:
            if name not in file_pd.columns:
                raise ValueError(f"The variable {name} is not in the txt file. \n {file}")

        select_pd = file_pd[varnames2load]
        select_pd.loc[:, "Time"] = pd.to_datetime(select_pd["Time"])
        data_all.append(select_pd)

    data_all = pd.concat(data_all, ignore_index=True)
    final_dict = data_all.to_dict(orient="list")
    for key in final_dict.keys():
        final_dict[key] = np.array(final_dict[key])

    if notplot:
        return final_dict
    else:
        for key in final_dict.keys():
            if key == "Time":
                continue
            store_data(prefix + key + suffix, data={"x": final_dict["Time"], "y": final_dict[key]})
        return list(final_dict.keys())


def magephem_read_h5(out_files, varnames=None, notplot=False, prefix="", suffix=""):
    """
    Read the h5 file and return the data as a dictionary or save to tplot variables.
    """

    if (varnames is None) or (len(varnames) == 0):
        varnames = ["IsoTime", "Lsimple", "Lm_eq", "CDMAG_MLAT", "CDMAG_MLON", "CDMAG_MLT"]

    data_all = []

    for ifile, file in enumerate(out_files):
        # open the h5 file
        f = h5py.File(file, "r")
        # ignore the foot of data
        date_array = np.char.decode(np.array(f.get("IsoTime")))
        time_dt64 = pd.to_datetime(date_array)
        file_pd = pd.DataFrame(columns=varnames)
        file_pd["IsoTime"] = time_dt64
        for name in varnames:
            if name == "IsoTime":
                continue
            if name not in h5_mag_name:
                raise ValueError(f"The variable {name} is not in the h5 file. \n {h5_mag_name}")
            else:
                data_name = np.array(f.get(name))
                file_pd[name] = data_name

        # If the time in the last row is 00:00:00, which duplicates the start of the next day, remove the last row
        if (ifile != len(out_files) - 1) and ("00:00:00" in date_array[-1]):
            file_pd = file_pd.iloc[:-1, :]

        data_all.append(file_pd)

    data_all = pd.concat(data_all, ignore_index=True)
    data_all = data_all.rename(columns={"IsoTime": "Time"})
    final_dict = data_all.to_dict(orient="list")

    for key in final_dict.keys():
        final_dict[key] = np.array(final_dict[key])

    if notplot:
        return final_dict
    else:
        for key in final_dict.keys():
            if key == "Time":
                continue
            store_data(prefix + key + suffix, data={"x": final_dict["Time"], "y": final_dict[key]})
        return list(final_dict.keys())



import unittest
from pyspedas import CDAWeb
from pytplot import data_exists, del_data, get_data, time_double, tplot_names, tplot


class CDAWebTests(unittest.TestCase):
    def test_get_observatories(self):
        cdaweb_obj = CDAWeb()
        obslist = cdaweb_obj.get_observatories()
        self.assertTrue('ARTEMIS' in obslist)

    def test_get_instruments(self):
        cdaweb_obj = CDAWeb()
        inst_list = cdaweb_obj.get_instruments()
        self.assertTrue('Magnetic Fields (space)' in inst_list)

    def test_get_datasets(self):
        cdaweb_obj = CDAWeb()
        ds_list = cdaweb_obj.get_datasets(['Voyager'], ['Magnetic Fields (space)'])
        self.assertTrue('VOYAGER2_2S_MAG (1977-08-24 to 1991-01-01)' in ds_list)

    def test_get_filenames(self):
        cdaweb_obj = CDAWeb()
        urllist = cdaweb_obj.get_filenames(['VOYAGER2_2S_MAG (1977-08-24 to 1991-01-01)'],
                                           "1979-01-11", "1979-01-12")
        self.assertTrue(any(["voyager2_2s_mag_19790111_v01.cdf" in item for item in urllist]))

    def test_load_merge(self):
        del_data('*')
        # Create the CDAWeb interface object
        cdaweb_obj = CDAWeb()

        # We'll pick one of available data sets and load it into tplot variables
        dataset = 'VOYAGER2_COHO1HR_MERGED_MAG_PLASMA'
        start_time = '2009-01-01 00:00:00'
        end_time = '2010-01-01 00:00:00'

        # Get the URLs for the available data in this time range
        urllist = cdaweb_obj.get_filenames([dataset], start_time, end_time)

        # Download the data and load as tplot variables.  Setting a prefix
        # is useful if you want to work with both Voyager 1 and Voyager 2
        # data; the variable names in the archived data are the same for both
        # spacecraft.

        cdaweb_obj.cda_download(urllist, "cdaweb/", prefix='v2_')
        dat = get_data('v2_protonDensity')
        # Files are monthly; we should have over a year of data.
        self.assertTrue((dat.times[-1] - dat.times[0])/86400.0 > 395.0)

    def test_load_icon_netcdf(self):
        del_data('*')
        # Create the CDAWeb interface object
        import pyspedas

        cdaweb_obj = pyspedas.CDAWeb()

        time0 = '2021-01-15T14:05:52'
        time1 = '2021-01-15T15:08:57'

        urllist = cdaweb_obj.get_filenames(['ICON_L2-2_MIGHTI_VECTOR-WIND-GREEN (2019-12-06 to 2022-11-25)'], time0,
                                           time1)
        result = cdaweb_obj.cda_download(urllist, "cdaweb/")
        self.assertTrue(data_exists('ICON_L22_Fringe_Amplitude'))

    def test_load_icon_netcdf_default_dir(self):
        del_data('*')
        # Create the CDAWeb interface object
        import pyspedas

        cdaweb_obj = pyspedas.CDAWeb()

        time0 = '2021-01-15T14:05:52'
        time1 = '2021-01-15T15:08:57'

        urllist = cdaweb_obj.get_filenames(['ICON_L2-2_MIGHTI_VECTOR-WIND-GREEN (2019-12-06 to 2022-11-25)'], time0,
                                           time1)
        result = cdaweb_obj.cda_download(urllist)
        self.assertTrue(data_exists('ICON_L22_Fringe_Amplitude'))

    def test_load_goes_netcdf(self):
        del_data('*')
        # Create the CDAWeb interface object
        import pyspedas

        cdaweb_obj = pyspedas.CDAWeb()

        time0 = '2021-01-15T14:05:52'
        time1 = '2021-01-15T15:08:57'

        urllist = cdaweb_obj.get_filenames(['DN_MAGN-L2-HIRES_G17 (2018-08-01 to 2023-03-14)'], time0,
                                           time1)
        result = cdaweb_obj.cda_download(urllist, "cdaweb/")
        self.assertTrue(data_exists('b_gse'))


    def test_load_time_clip(self):
        del_data('*')
        cdaweb_obj = CDAWeb()

        ds_list = cdaweb_obj.get_datasets(['THEMIS'], ['Magnetic Fields (space)'])

        # End date of THEMIS mag data is continuously updated, so we have to search the available datesets
        # rather than hardcoding it

        dataset = next((item for item in ds_list if 'THA_L2_FIT' in item), None)

        self.assertTrue(dataset is not None)

        start_time = '2010-01-01 00:00:00'
        end_time = '2010-01-01 12:00:00'

        # Get the URLs for the available data in this time range
        urllist = cdaweb_obj.get_filenames([dataset], start_time, end_time)
        cdaweb_obj.cda_download(urllist, "cdaweb/", trange=[start_time, end_time], time_clip=True)

        # Verify that the times were clipped
        # Note that tplot variables with no data in the requested interval will NOT be clipped (deleted).
        dat = get_data('tha_fgs_dsl')
        self.assertTrue(dat.times[0] >= time_double(start_time))
        self.assertTrue(dat.times[-1] <= time_double(end_time))

    def test_load_time_clip_no_trange(self):
        del_data('*')
        cdaweb_obj = CDAWeb()

        ds_list = cdaweb_obj.get_datasets(['THEMIS'], ['Magnetic Fields (space)'])

        # End date of THEMIS mag data is continuously updated, so we have to search the available datesets
        # rather than hardcoding it

        dataset = next((item for item in ds_list if 'THA_L2_FIT' in item), None)

        self.assertTrue(dataset is not None)

        start_time = '2010-01-01 00:00:00'
        end_time = '2010-01-01 12:00:00'

        # Get the URLs for the available data in this time range
        urllist = cdaweb_obj.get_filenames([dataset], start_time, end_time)
        with self.assertLogs(level='WARNING') as logs:
            cdaweb_obj.cda_download(urllist, "cdaweb/", time_clip=True)
        got_trange_warning = False
        for o in logs.output:
            if "No trange specified" in o:
                got_trange_warning = True
        self.assertTrue(got_trange_warning)

    def test_load_time_clip_empty_trange(self):
        del_data('*')
        cdaweb_obj = CDAWeb()

        ds_list = cdaweb_obj.get_datasets(['THEMIS'], ['Magnetic Fields (space)'])

        # End date of THEMIS mag data is continuously updated, so we have to search the available datesets
        # rather than hardcoding it

        dataset = next((item for item in ds_list if 'THA_L2_FIT' in item), None)

        self.assertTrue(dataset is not None)

        start_time = '2010-01-01 00:00:00'
        end_time = '2010-01-01 12:00:00'

        # Get the URLs for the available data in this time range
        urllist = cdaweb_obj.get_filenames([dataset], start_time, end_time)
        with self.assertLogs(level='WARNING') as logs:
            cdaweb_obj.cda_download(urllist, "cdaweb/", trange=[start_time, start_time],  time_clip=True)
        got_trange_warning = False
        for o in logs.output:
            if "equal" in o:
                got_trange_warning = True
        self.assertTrue(got_trange_warning)

    def test_load_time_clip_backward_trange(self):
        del_data('*')
        cdaweb_obj = CDAWeb()

        ds_list = cdaweb_obj.get_datasets(['THEMIS'], ['Magnetic Fields (space)'])

        # End date of THEMIS mag data is continuously updated, so we have to search the available datesets
        # rather than hardcoding it

        dataset = next((item for item in ds_list if 'THA_L2_FIT' in item), None)

        self.assertTrue(dataset is not None)

        start_time = '2010-01-01 00:00:00'
        end_time = '2010-01-01 12:00:00'

        # Get the URLs for the available data in this time range
        urllist = cdaweb_obj.get_filenames([dataset], start_time, end_time)
        with self.assertLogs(level='WARNING') as logs:
            cdaweb_obj.cda_download(urllist, "cdaweb/", trange=[end_time, start_time],  time_clip=True)
        got_trange_warning = False
        for o in logs.output:
            if "out of order" in o:
                got_trange_warning = True
        self.assertTrue(got_trange_warning)

    def test_load_data_prefix(self):
        del_data('*')
        cdaweb_obj = CDAWeb()
        urllist = cdaweb_obj.get_filenames(['VOYAGER2_2S_MAG (1977-08-24 to 1991-01-01)'],
                                           "1979-01-11", "1979-01-12")
        cdaweb_obj.cda_download(urllist, "cdaweb/", prefix='v2_')
        self.assertTrue(data_exists('v2_B1'))

    def test_load_icon_netcdf_alt_urls(self):
        del_data('*')
        # Create the CDAWeb interface object
        import pyspedas
        from pyspedas.cdagui_tools.config import CONFIG
        # Save the URL settings from the current configuration
        save_cdas_endpoint = CONFIG['cdas_endpoint']
        save_remote_data_dir = CONFIG['remote_data_dir']

        # Replace the standard URLs with http rather than https and see if it still works
        alternate_endpoint = 'http://cdaweb.gsfc.nasa.gov/WS/cdasr/1/dataviews/sp_phys/'
        alternate_download_url = 'http://cdaweb.gsfc.nasa.gov/sp_phys/data'

        CONFIG['cdas_endpoint'] = alternate_endpoint
        CONFIG['remote_data_dir'] = alternate_download_url

        cdaweb_obj = pyspedas.CDAWeb()

        # I'm not sure if it's possible to get the cdasws library to issue a log message with the endpoint URL that
        # was actually used.  So we'll go ahead and peek at the internals, just to see if the object got initialized as expected.
        endpoint_used = cdaweb_obj.cdas._endpoint

        # Note that the time range is different from the other ICON test, to avoid tests interfering with each other
        time0 = '2021-02-15T14:05:52'
        time1 = '2021-02-15T15:08:57'

        urllist = cdaweb_obj.get_filenames(['ICON_L2-2_MIGHTI_VECTOR-WIND-GREEN (2019-12-06 to 2022-11-25)'], time0,
                                           time1)
        result = cdaweb_obj.cda_download(urllist, "cdaweb/")

        # Restore prior configured URLs
        CONFIG['cdas_endpoint'] = save_cdas_endpoint
        CONFIG['remote_data_dir'] = save_remote_data_dir

        # Check that the expected endpoint was used
        self.assertEqual(endpoint_used, alternate_endpoint)
        # Check that the expected download URL was used
        self.assertTrue(alternate_download_url in urllist[0])
        # Check that data was successfully loaded
        self.assertTrue(data_exists('ICON_L22_Fringe_Amplitude'))


if __name__ == '__main__':
    unittest.main()
